#pragma once
#include <cstdint>
#include <vector>

namespace pythonLevenshteinIssue9 {
extern std::vector<uint8_t> example1;
extern std::vector<uint8_t> example2;
} // namespace pythonLevenshteinIssue9
